/*----------------------------------------------------------------------------*
 *  vrml.y
 *           
 * Project: SICS DIVE
 * Copyright: SICS
 * Implemented by: Emmanuel Frcon and Olof Hagsand
 *
 *  This file contains yacc declarations a VRML Parser.
 *
 *---------------------------------------------------------------------------*/

%union {
  int intval;
  float floatval;
  double doubleval;
  char *string;
  void *any_type;
}


%token <intval> NUMBER 
%token <floatval> FLOAT
%token <string> VRMLSTRING VRMLWORD

%token ASCIITEXT CONE COORDINATE3 CUBE CYLINDER DIRECTIONALLIGHT FONTSTYLE
%token GROUP INDEXEDFACESET INDEXEDLINESET INFO LOD MATERIAL MATERIALBINDING
%token LEVELOFDETAIL MATRIXTRANSFORM
%token NORMAL NORMALBINDING ORTHOGRAPHICCAMERA PERSPECTIVECAMERA
%token POINTLIGHT POINTSET ROTATION SCALE SEPARATOR SHAPEHINTS SPHERE
%token SPOTLIGHT SWITCH TEXTURE2 TEXTURE2TRANSFORM TEXTURECOORDINATE2
%token TRANSFORM TRANSFORMSEPARATOR TRANSLATION WWWANCHOR WWWINLINE

%token PARTS BOTTOMRADIUS HEIGHT POINT WIDTH DEPTH STRING SPACING
%token JUSTIFICATION RADIUS ON INTENSITY COLOR DIRECTION SIZE FAMILY STYLE
%token COORDINDEX MATERIALINDEX NORMALINDEX TEXTURECOORDINDEX RANGE
%token SCREENAREA TRANSLATION_F
%token CENTER AMBIENTCOLOR DIFFUSECOLOR SPECULARCOLOR EMISSIVECOLOR SHININESS
%token TRANSPARENCY VALUE MATRIX VECTOR POSITION ORIENTATION FOCALDISTANCE
%token HEIGHTANGLE LOCATION STARTINDEX NUMPOINTS ROTATION_F SCALEFACTOR
%token RENDERCULLING VERTEXORDERING SHAPETYPE FACETYPE CREASEANGLE DROPOFFRATE
%token CUTOFFANGLE WHICHCHILD FILENAME IMAGE WRAPS WRAPT SCALEORIENTATION
%token NAME DESCRIPTION MAP BBOXSIZE BBOXCENTER

%token DEF USE

%token FIELDS UNKNOWNNODE

%type <floatval> SFFloat
%type <intval> SFBitMask bitmasklist SFBool SFEnum SFLong
%type <string> SFString  

%start vrml

%{

#define YYMAXDEPTH 100000
#include <stdio.h>




/*==========================================================================
  Local variables
  ==========================================================================*/
static char rcsid[] = "$Id: file.y,v 1.11 1995/05/26 12:13:20 olof Exp $";

/*==========================================================================
  Exported variables
  ==========================================================================*/


%} 

%%
/*****************************************************************************/
/* syntactical definitions
/*****************************************************************************/

vrml		: Nodes {YYACCEPT;}
		| error {YYABORT;}
		;

Nodes		: Node Nodes
		|
		;

Node		: NodeSpec NodeBody 
		| USE VRMLWORD
		;

NodeSpec	: DEF VRMLWORD
		|
		;

NodeBody	: AsciiText
		| Cone
		| Coordinate3
		| Cube
		| Cylinder
		| Dlight
		| FontStyle
		| Group
		| IdxFaceSet
		| IdxLineSet
		| Info
		| Lod
		| LevelOfDetail
		| Material
		| MaterialBinding
		| MatrixTransform
		| Normal
		| NormalBinding
		| OrthographicCamera
		| PerspectiveCamera
		| PointLight
		| PointSet
		| Rotation
		| Scale
		| Separator
		| ShapeHints
		| Sphere
		| SpotLight
		| Switch
		| Texture2
		| Tex2Transform
		| TexCoordinate2
		| Transform
		| TransSeparator
		| Translation
		| WWWAnchor
		| WWWInline
		| UnknownNode
		;

/* Field specification */
FieldSpecs	: FIELDS
		|
		;

/* Unknown node */
UnknownNode	: UNKNOWNNODE
		;

/* Ascii Text */
asciitextfields	: asciitextfield asciitextfields
		|
		;

asciitextfield	: STRING	MFString
		| SPACING	SFFloat
		| JUSTIFICATION	SFEnum
		| WIDTH		MFFloat
		;
		
AsciiText	: ASCIITEXT '{' FieldSpecs asciitextfields '}'
		;

/* Cones */
conefields	: conefield conefields
		|
		;

conefield	: PARTS		SFBitMask
		| BOTTOMRADIUS	SFFloat
		| HEIGHT	SFFloat
		;

Cone		: CONE '{' FieldSpecs conefields '}'
		;


/* Coordinate 3 */
coordinatefield :  POINT MFVec3f
		|
		;

Coordinate3	: COORDINATE3 '{' FieldSpecs coordinatefield '}'
		;

/* Cubes */
cubefields	: cubefield cubefields
		|
		;

cubefield	: WIDTH		SFFloat
		| HEIGHT	SFFloat
		| DEPTH		SFFloat
		;

Cube		: CUBE '{' FieldSpecs cubefields '}'
		;


/* Cylinder */
cylinderfields	: cylinderfield cylinderfields
		|
		;

cylinderfield	: PARTS		SFBitMask
		| RADIUS	SFFloat
		| HEIGHT	SFFloat
		;

Cylinder	: CYLINDER '{' FieldSpecs cylinderfields '}'
		;

/* Directional Light */
dlightfields	: dlightfield dlightfields
		|
		;

dlightfield	: ON		SFBool
		| INTENSITY	SFFloat
		| COLOR		SFColor
		| DIRECTION	SFVec3f
		;

Dlight		: DIRECTIONALLIGHT '{' FieldSpecs dlightfields '}'
		;

/* Font Style */
fstylefields	: fstylefield fstylefields
		|
		;

fstylefield	: SIZE		SFFloat
		| FAMILY	SFEnum
		| STYLE		SFBitMask
		;

FontStyle	: FONTSTYLE '{' FieldSpecs fstylefields '}'
		;

/* Groups */
Group		: GROUP '{' Nodes '}'
		;

/* Indexed Face Set */
ifacesetfields	: ifacesetfield ifacesetfields
		|
		;

ifacesetfield	: COORDINDEX		MFLong
		| MATERIALINDEX		MFLong
		| NORMALINDEX		MFLong
		| TEXTURECOORDINDEX	MFLong
		;

IdxFaceSet	: INDEXEDFACESET '{' FieldSpecs ifacesetfields '}'
		;

/* Indexed Line Set */
ilinesetfields	: ilinesetfield ilinesetfields
		|
		;

ilinesetfield	: COORDINDEX		MFLong
		| MATERIALINDEX		MFLong
		| NORMALINDEX		MFLong
		| TEXTURECOORDINDEX	MFLong
		;

IdxLineSet	: INDEXEDLINESET '{' FieldSpecs ilinesetfields '}'
		;

/* Info */
infofield	: STRING	SFString
		|
		;

Info		: INFO '{' FieldSpecs infofield '}'
		;

/* Level of Detail 
   LOD comes from the new VRML draft, LEVELOFDETAIL from the old one !! */
lodfields	: lodfield lodfields
		|
		;

lodfield	: RANGE		MFFloat
		| CENTER	SFVec3f
		;

Lod		: LOD '{' FieldSpecs lodfields Nodes '}'
		;

oldlodfield	: SCREENAREA	MFFloat
		|
		;

LevelOfDetail	: LEVELOFDETAIL '{' FieldSpecs oldlodfield Nodes '}'
		;

/* Material */
materialfields	: materialfield materialfields
		|
		;

materialfield	: AMBIENTCOLOR		MFColor
		| DIFFUSECOLOR		MFColor
		| SPECULARCOLOR		MFColor
		| EMISSIVECOLOR		MFColor
		| SHININESS		MFFloat
		| TRANSPARENCY		MFFloat
		;

Material	: MATERIAL '{' FieldSpecs materialfields '}'
		;

/* Material Binding */
mbindingfields	: mbindingfield mbindingfields
		|
		;

mbindingfield	: VALUE			SFEnum
		;

MaterialBinding	: MATERIALBINDING '{' FieldSpecs mbindingfields '}'
		;

/* Matrix Transform */
mxtransformfields	: mxtransformfield mxtransformfields
			|
			;

mxtransformfield	: MATRIX	SFMatrix
			;

MatrixTransform	: MATRIXTRANSFORM '{' FieldSpecs mxtransformfields '}'
		;

/* Normal */
normalfields	: normalfield normalfields
		|
		;

normalfield	: VECTOR	MFVec3f
		;

Normal		: NORMAL '{' FieldSpecs normalfields '}'
		;

/* Normal Binding */
nbindingfields	: nbindingfield nbindingfields
		|
		;

nbindingfield	: VALUE			SFEnum
		;

NormalBinding	: NORMALBINDING '{' FieldSpecs nbindingfields '}'
		;

/* Orthographic camera */
ocamerafields	: ocamerafield ocamerafields
		|
		;

ocamerafield	: POSITION	SFVec3f
		| ORIENTATION	SFRotation
		| FOCALDISTANCE	SFFloat
		| HEIGHT	SFFloat
		;

OrthographicCamera	: ORTHOGRAPHICCAMERA '{' FieldSpecs ocamerafields '}'
			;

/* Perspective camera */
pcamerafields	: pcamerafield pcamerafields
		|
		;

pcamerafield	: POSITION	SFVec3f
		| ORIENTATION	SFRotation
		| FOCALDISTANCE	SFFloat
		| HEIGHTANGLE	SFFloat
		;

PerspectiveCamera	: PERSPECTIVECAMERA '{' FieldSpecs pcamerafields '}'
			;

/* Point Light */
plightfields	: plightfield plightfields
		|
		;

plightfield	: ON		SFBool
		| INTENSITY	SFFloat
		| COLOR		SFColor
		| LOCATION	SFVec3f
		;	

PointLight	: POINTLIGHT '{' FieldSpecs plightfields '}'
		;

/* Point Set */
psetfields	: psetfield psetfields
		|
		;

psetfield	: STARTINDEX	SFLong
		| NUMPOINTS	SFLong
		;

PointSet	: POINTSET '{' FieldSpecs psetfields '}'
		;

/* Rotation */
rotationfields	: rotationfield rotationfields
		|
		;

rotationfield	: ROTATION_F	SFRotation
		;

Rotation	: ROTATION '{' FieldSpecs rotationfields '}'
		;

/* Scale */
scalefields	: scalefield scalefields
		|
		;

scalefield	: SCALEFACTOR	SFVec3f
		;

Scale		: SCALE '{' FieldSpecs scalefields '}'
		;

/* Separator */
separatorfields	: separatorfield separatorfields
		|
		;

separatorfield	: RENDERCULLING	SFEnum
		;

Separator	: SEPARATOR '{' FieldSpecs separatorfields Nodes '}'
		;

/* Shape Hints */
shapehintsfields	: shapehintsfield shapehintsfields
			|
			;

shapehintsfield	: VERTEXORDERING	SFEnum
		| SHAPETYPE		SFEnum
		| FACETYPE		SFEnum
		| CREASEANGLE		SFFloat
		;

ShapeHints	: SHAPEHINTS '{' FieldSpecs shapehintsfields '}'
		;

/* Sphere */
spherefields	: spherefield spherefields
		|
		;

spherefield	: RADIUS	SFFloat
		;

Sphere		: SPHERE '{' FieldSpecs spherefields '}'
		;

/* Spotlight */
spotlightfields	: spotlightfield spotlightfields
		|
		;

spotlightfield	: ON		SFBool
		| INTENSITY	SFFloat
		| COLOR		SFColor
		| LOCATION	SFVec3f
		| DIRECTION	SFVec3f
		| DROPOFFRATE	SFFloat
		| CUTOFFANGLE	SFFloat
		;

SpotLight	: SPOTLIGHT '{' FieldSpecs spotlightfields '}'
		;

/* Switch */
switchfields	: switchfield switchfields
		|
		;

switchfield	: WHICHCHILD	SFLong
		;

Switch		: SWITCH '{' FieldSpecs switchfields Nodes '}'
		;

/* Texture2 */
texture2fields	: texture2field texture2fields
		|
		;

texture2field	: FILENAME	SFString
		| IMAGE		SFImage
		| WRAPS		SFEnum
		| WRAPT		SFEnum
		;

Texture2	: TEXTURE2 '{' FieldSpecs texture2fields '}'
		;

/* Texture transformation */
tex2transfields	: tex2transfield tex2transfields
		|
		;

tex2transfield	: TRANSLATION_F	SFVec2f
		| ROTATION_F	SFFloat
		| SCALEFACTOR	SFVec2f
		| CENTER	SFVec2f
		;

Tex2Transform	: TEXTURE2TRANSFORM '{' FieldSpecs tex2transfields '}'
		;

/* Texcoord2 */
texcoord2fields	: texcoord2field texcoord2fields
		|
		;

texcoord2field	: POINT		MFVec2f
		;

TexCoordinate2	: TEXTURECOORDINATE2 '{' FieldSpecs texcoord2fields '}'
		;

/* Transformation */
transformfields	: transformfield transformfields
		|
		;

transformfield	: TRANSLATION_F		SFVec3f
		| ROTATION_F		SFRotation
		| SCALEFACTOR		SFVec3f
		| SCALEORIENTATION	SFRotation
		| CENTER		SFVec3f
		;

Transform	: TRANSFORM '{' FieldSpecs transformfields '}'
		;

/* Transform Separator */
TransSeparator	: TRANSFORMSEPARATOR '{' FieldSpecs Nodes '}'
		;

/* Translation */
translationfields	: translationfield translationfields
			|
			;

translationfield	: TRANSLATION_F	SFVec3f
			;

Translation	: TRANSLATION '{' FieldSpecs translationfields '}'
		;

/* WWWAnchor */
WWWanchorfields	: WWWanchorfield WWWanchorfields
		|
		;

WWWanchorfield	: NAME		SFString
		| DESCRIPTION	SFString
		| MAP		SFEnum
		;

WWWAnchor	: WWWANCHOR '{' FieldSpecs WWWanchorfields Nodes '}'
		;

/* WWWinline */
WWWinlinefields	: WWWinlinefield WWWinlinefields
		|
		;

WWWinlinefield	: NAME		SFString
		| BBOXSIZE	SFVec3f
		| BBOXCENTER	SFVec3f
		;

WWWInline	: WWWINLINE '{' FieldSpecs WWWinlinefields '}'
		;





/* ========== Classes of Fields ========== */

/* Single Value Fields */
bitmasklist	: NUMBER 
		| bitmasklist '|' NUMBER {$$ = ($1 | $3);}
		;

SFBitMask	: NUMBER
		| '(' bitmasklist ')' {$$ = $2;}
		;

SFBool		: NUMBER
		;

SFColor		: SFFloat SFFloat SFFloat {}
		;

SFEnum		: NUMBER
		;

SFFloat		: FLOAT
		| NUMBER {$$ = (float)$1;}
		;

imglist		: NUMBER imglist {}
		|
		;

SFImage		: NUMBER NUMBER NUMBER imglist {}
		;

SFLong		: NUMBER
		;

SFMatrix	: SFFloat SFFloat SFFloat SFFloat
		  SFFloat SFFloat SFFloat SFFloat
		  SFFloat SFFloat SFFloat SFFloat
		  SFFloat SFFloat SFFloat SFFloat {}
		;

SFRotation	: SFFloat SFFloat SFFloat SFFloat {}
		;

SFString	: VRMLSTRING
		;

SFVec2f		: SFFloat SFFloat {}
		;

SFVec3f		: SFFloat SFFloat SFFloat {}
		;

/* Multiple Value Field */
colorlist	: SFColor
		| SFColor ',' colorlist
		;

MFColor		: SFColor 
		| '[' colorlist ']'
		;

longlist	: SFLong {}
		| SFLong ',' longlist {}
		;

MFLong		: SFLong {}
		| '[' longlist ']' {}
		; 

floatlist	: SFFloat {}
		| SFFloat ',' floatlist {}
		;

MFFloat		: SFFloat {}
		| '[' floatlist ']' {}
		; 

stringlist	: SFString {}
		| SFString ',' stringlist {}
		;

MFString	: SFString {}
		| '[' stringlist ']' {}
		; 

vec2flist	: SFVec2f
		| SFVec2f ',' vec2flist
		;

MFVec2f		: SFVec2f
		| '[' vec2flist ']'
		; 

vec3flist	: SFVec3f
		| SFVec3f ',' vec3flist
		;

MFVec3f		: SFVec3f
		| '[' vec3flist ']'
		; 

%%

yyerror(char *s) 
{ 
  extern  char    yytext[];
  int linenum;
  char *ff;

  linenum = vrml_filel_linenum(&ff);
  fprintf(stderr, "%s:%d: Error:\n%s at or before '%s'\n\n\n", 
	  ff,linenum, (char *)s, *yytext=='\n'? "\\n": yytext); 
  return;
}
